"""
Run makecldf command of a dataset
"""
from cldfbench.cli_util import with_dataset, get_dataset

from pylexibank.cli_util import add_catalogs, add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    add_catalogs(parser, with_clts=True)
    parser.add_argument('--verbose', action='store_true')


def run(args):
    dataset = get_dataset(args)
    dataset.concepticon = args.concepticon.api
    dataset.glottolog = args.glottolog.api
    with_dataset(args, 'makecldf', dataset=dataset)
    if not dataset.cldf_dir.joinpath('sources.bib').exists():
        raise ValueError(
            'The dataset has no sources at {0}'.format(dataset.cldf_dir.joinpath('sources.bib')))



import re
from collections import OrderedDict
from pathlib import Path

from termcolor import colored
from tqdm import tqdm

from clldutils.misc import slug
from clldutils.badge import Colors, badge
from clldutils import jsonlib
from cldfbench.datadir import get_url
from pycldf.sources import Source, Reference

__all__ = ['progressbar', 'getEvoBibAsBibtex']
YEAR_PATTERN = re.compile('\s+\(?(?P<year>[1-9][0-9]{3}(-[0-9]+)?)(\)|\.)')


def progressbar(iterable=None, **kw):
    kw.setdefault('leave', False)
    kw.setdefault('desc', 'cldfbench')
    return tqdm(iterable=iterable, **kw)


# backwards compat:
pb = progressbar


def split_by_year(s):
    match = YEAR_PATTERN.search(s)
    if match:
        return s[:match.start()].strip(), match.group('year'), s[match.end():].strip()
    return None, None, s


def get_reference(author, year, title, pages, sources, id_=None, genre='misc'):
    kw = {'title': title}
    id_ = id_ or None
    if author and year:
        id_ = id_ or slug(author + year)
        kw.update(author=author, year=year)
    elif title:
        id_ = id_ or slug(title)

    if not id_:
        return

    source = sources.get(id_)
    if source is None:
        sources[id_] = source = Source(genre, id_, **kw)

    return Reference(source, pages)


def get_badge(ratio, name):
    if ratio >= 0.99:
        color = Colors.brightgreen
    elif ratio >= 0.9:
        color = 'green'
    elif ratio >= 0.8:
        color = Colors.yellowgreen
    elif ratio >= 0.7:
        color = Colors.yellow
    elif ratio >= 0.6:
        color = Colors.orange
    else:
        color = Colors.red
    ratio = int(round(ratio * 100))
    return badge(name, '%s%%' % ratio, color, label="{0}: {1}%".format(name, ratio))


def sorted_obj(obj):
    res = obj
    if isinstance(obj, dict):
        res = OrderedDict()
        obj.pop(None, None)
        for k, v in sorted(obj.items()):
            res[k] = sorted_obj(v)
    elif isinstance(obj, (list, set)):
        res = [sorted_obj(v) for v in obj]
    return res


def log_dump(fname, log=None):
    if log:
        log.info('file written: {0}'.format(colored(fname.as_posix(), 'green')))


def jsondump(obj, fname, log=None):
    fname = Path(fname)
    if fname.exists():
        d = jsonlib.load(fname)
        d.update(obj)
        obj = d
    jsonlib.dump(sorted_obj(obj), fname, indent=4)
    log_dump(fname, log=log)
    return obj


def getEvoBibAsBibtex(*keys, **kw):
    """Download bibtex format and parse it from EvoBib"""
    res = []
    for key in keys:
        bib = get_url(
            "http://bibliography.lingpy.org/raw.php?key=" + key,
            log=kw.get('log')).text
        try:
            res.append('@' + bib.split('@')[1].split('</pre>')[0])
        except IndexError:  # pragma: no cover
            raise KeyError('Missing entry {0} in evobib'.format(key))
    return '\n\n'.join(res)



import re

import attr
from clldutils.misc import slug, nfilter
from pycldf.sources import Source

from pylexibank.util import pb
from pylexibank import Dataset, Language

BASE_URL = "https://abvd.shh.mpg.de"
URL = BASE_URL + "/utils/save/?type=xml&section=%s&language=%d"
INVALID_LANGUAGE_IDS = {
    'austronesian': [261],  # Duplicate West Futuna list
}


@attr.s
class BVDLanguage(Language):
    author = attr.ib(default=None)
    url = attr.ib(default=None)
    typedby = attr.ib(default=None)
    checkedby = attr.ib(default=None)
    notes = attr.ib(default=None)


class BVD(Dataset):
    SECTION = None
    language_class = BVDLanguage
    cognate_pattern = re.compile('\s*(?P<id>([A-z]?[0-9]+|[A-Z]))\s*(?P<doubt>\?+)?\s*$')

    def iter_wordlists(self, language_map, log):
        for xml in pb(
                sorted(self.raw_dir.glob('*.xml'), key=lambda p: int(p.stem)), desc='xml-to-wl'):
            wl = Wordlist(self, xml, log)
            if not wl.language.glottocode:
                if wl.language.id in language_map:
                    wl.language.glottocode = language_map[wl.language.id]
                else:  # pragma: no cover
                    self.unmapped.add_language(
                        ID=wl.language.id,
                        Name=wl.language.name,
                        ISO639P3code=wl.language.iso
                    )
            yield wl

    def cmd_download(self, **kw):  # pragma: no cover
        assert self.SECTION in ['austronesian', 'mayan', 'utoaztecan']
        self.log.info('ABVD section set to %s' % self.SECTION)
        for fname in self.raw_dir.iterdir():
            fname.unlink()
        language_ids = [
            i for i in range(1, 2000)
            if i not in INVALID_LANGUAGE_IDS.get(self.SECTION, [])]

        for lid in language_ids:
            if not self.get_data(lid):
                self.log.warn("No content for %s %d. Ending." % (self.SECTION, lid))
                break

    def get_data(self, lid):  # pragma: no cover
        fname = self.raw_dir.download(URL % (self.SECTION, lid), '%s.xml' % lid, log=self.log)
        if fname.stat().st_size == 0:
            fname.unlink()
            return False
        return True


class XmlElement(object):
    __attr_map__ = []

    def __init__(self, e):
        self.id = '%s' % int(e.find('id').text)
        self.e = e
        for spec in self.__attr_map__:
            if len(spec) == 2:
                attr, nattr = spec
                conv = None
            elif len(spec) == 3:
                attr, nattr, conv = spec
            else:
                raise ValueError(spec)  # pragma: no cover
            nattr = nattr or attr
            ee = e.find(attr)
            if ee is not None:
                text = e.find(attr).text
                if text and not isinstance(text, str):
                    text = text.decode('utf8')  # pragma: no cover
            else:
                text = ''
            if text:
                text = text.strip()

            if text and conv:
                text = conv(text)

            setattr(self, nattr, text or None)


class Language(XmlElement):
    __attr_map__ = [
        ('author', ''),
        ('language', 'name'),
        ('silcode', 'iso'),
        ('glottocode', ''),
        ('notes', ''),
        ('problems', ''),
        ('classification', ''),
        ('typedby', ''),
        ('checkedby', ''),
    ]

    def __init__(self, e):
        XmlElement.__init__(self, e)


class Entry(XmlElement):
    __attr_map__ = [
        ('id', '', lambda s: '%s' % int(s)),
        ('word_id', '', lambda s: '%s' % int(s)),
        ('word', ''),
        ('item', 'name'),
        ('annotation', 'comment'),
        ('loan', ''),
        ('cognacy', ''),
    ]

    def __init__(self, e, section):
        self.section = section
        self.rowids = []
        XmlElement.__init__(self, e)

    @property
    def cognates(self):
        # We split into single cognateset IDs on comma, slash and dot.
        return nfilter(re.split('[,/.]', self.cognacy or ''))


class Wordlist(object):
    def __init__(self, dataset, path, log):
        self.dataset = dataset
        self.log = log
        e = dataset.raw_dir.read_xml(path.name)
        self.section = dataset.SECTION
        records = list(e.findall('./record'))
        self.language = Language(records[0])
        self.entries = [Entry(r, self.section) for r in records[1:] if self.is_entry(r)]

    @staticmethod
    def is_entry(r):
        return getattr(r.find('id'), 'text', None) \
            and getattr(r.find('item'), 'text', None)

    def url(self, path):
        return '%s/bantu/%s/%s' % (BASE_URL, self.section, path)

    @property
    def name(self):
        return '%s - %s - %s' % (self.section, self.language.name, self.id)

    @property
    def id(self):
        return '%s-%s' % (self.section, self.language.id)

    def md(self):
        return dict(properties={
            k: getattr(self.language, k, None)
            for k in 'id name author notes problems typedby checkedby'.split()})

    def to_cldf(self, ds, concept_map, citekey=None, source=None, concept_key=None):
        if concept_key is None:
            def concept_key(entry):
                return entry.word_id

        ref = None
        if citekey and source:
            ref = citekey
            for r in ref.split(";"):
                for s in source:
                    if isinstance(s, Source):
                        ds.add_sources(s)
                    else:
                        ds.add_sources(Source('misc', r, title=s))

        ds.add_language(
            ID=self.language.id,
            Glottocode=self.language.glottocode,
            ISO639P3code=self.language.iso,
            Name=self.language.name,
            author=self.language.author,
            url=self.url('language.php?id=%s' % self.language.id),
            typedby=self.language.typedby,
            checkedby=self.language.checkedby,
            notes=self.language.notes,
        )

        for entry in self.entries:
            if entry.name is None or len(entry.name) == 0:  # skip empty entries
                continue  # pragma: no cover

            if entry.cognacy and (
                    's' == entry.cognacy.lower() or 'x' in entry.cognacy.lower()):
                # skip entries marked as incorrect word form due to semantics
                # (x = probably, s = definitely)
                continue  # pragma: no cover

            if not (citekey and source):
                src = entry.e.find('source')
                if (src is not None) and getattr(src, 'text'):
                    ref = slug(str(src.text))
                    ds.add_sources(Source('misc', ref, title=src.text))
            cid = concept_map.get(concept_key(entry))
            if not cid:
                self.dataset.unmapped.add_concept(ID=entry.word_id, Name=entry.word)

            ds.add_concept(ID=entry.word_id, Name=entry.word, Concepticon_ID=cid)
            lex = ds.add_forms_from_value(
                Language_ID=self.language.id,
                Parameter_ID=entry.word_id,
                Value=entry.name,
                Source=[ref],
                Cognacy=entry.cognacy,
                Comment=entry.comment or '',
                Loan=True if entry.loan and len(entry.loan) else False,
                Local_ID=entry.id,
            )
            if lex:
                lex = lex[0]
                for cognate_set_id in entry.cognates:
                    match = self.dataset.cognate_pattern.match(cognate_set_id)
                    if not match:  # pragma: no cover
                        self.log.warn('Invalid cognateset ID for entry {0}: {1}'.format(
                            entry.id, cognate_set_id))
                    else:
                        ds.add_cognate(
                            lexeme=lex,
                            Cognateset_ID=match.group('id'),
                            Doubt=bool(match.group('doubt')))

        return ds



"""
Load dataset(s) into the lexibank SQLite database.
"""
from cldfbench.cli_util import with_datasets

from pylexibank.cli_util import add_db, add_catalogs, get_db, add_dataset_spec


def register(parser):
    add_db(parser)
    add_dataset_spec(parser, multiple=True)
    add_catalogs(parser)


def run(args):
    db = get_db(args)
    with_datasets(args, db.load)
    db.load_concepticon_data(args.concepticon.api)
    db.load_glottolog_data(args.glottolog.api)



"""
Check forms against a dataset's orthography profile.
"""
from cldfbench.cli_util import with_dataset, add_catalog_spec

from pylexibank.cli_util import add_dataset_spec


def register(parser):
    add_dataset_spec(parser)
    add_catalog_spec(parser, 'clts')


def run(args):
    with_dataset(args, check_profile)


def check_profile(dataset, args):
    problems, visited = set(), set()
    for row in dataset.cldf_dir.read_csv('forms.csv', dicts=True):
        tokens = dataset.tokenizer(None, row['Form'], column='IPA') if \
            dataset.tokenizer else row['Segments'].split()
        for tk in set(tokens):
            if tk not in visited:
                visited.add(tk)
                if args.clts.api.bipa[tk].type == 'unknownsound':
                    problems.add(tk)
                    print('{0:5}\t{1:20}\t{2}'.format(tk, ' '.join(tokens), row['Form']))
    print('Found {0} errors in {1} segments.'.format(len(problems), len(visited)))



#
from pylexibank.dataset import *  # noqa: F401, F403
from pylexibank.models import *  # noqa: F401, F403
from pylexibank.forms import *  # noqa: F401, F403
from pylexibank.metadata import *  # noqa: F401, F403
from pylexibank.util import *  # noqa: F401, F403
from pylexibank.cldf import *  # noqa: F401, F403

__version__ = '2.1.0'



from zipfile import ZipFile

from clldutils.misc import lazyproperty
from pycldf import Wordlist
from pycldf.dataset import MD_SUFFIX

from pylexibank.dataset import Dataset


class CLLD(Dataset):
    __cldf_url__ = None

    def url(self, path='/'):
        return "http://{0}.clld.org{1}".format(self.id, path)

    def cmd_download(self, args):
        zname = '{0}.zip'.format(self.id)
        self.raw_dir.download(self.__cldf_url__, zname, log=args.log)
        archive = ZipFile(str(self.raw_dir / zname))
        archive.extractall(str(self.raw_dir))
        (self.raw_dir / zname).unlink()

    def add_sources(self, ds):
        ds.add_sources((self.raw_dir / 'sources.bib').read_text(encoding='utf8'))

    @lazyproperty
    def original_cldf(self):
        for p in self.raw_dir.iterdir():
            if p.name.endswith(MD_SUFFIX):
                return Wordlist.from_metadata(p)



import pathlib

from cldfbench.scaffold import Template

import pylexibank
from pylexibank.metadata import LexibankMetadata


_TEMPLATES_DIR = pathlib.Path(pylexibank.__file__).parent / 'dataset_templates'


class LexibankTemplate(Template):
    prefix = 'lexibank'
    package = pylexibank.__name__

    dirs = Template.dirs + [_TEMPLATES_DIR / 'lexibank_simple']
    metadata = LexibankMetadata


class LexibankCombinedTemplate(Template):
    prefix = 'lexibank'
    package = pylexibank.__name__

    dirs = Template.dirs + \
        [_TEMPLATES_DIR / 'lexibank_simple', _TEMPLATES_DIR / 'lexibank_combined']
    metadata = LexibankMetadata



import pathlib
import unicodedata

import attr

from csvw.dsv import reader
from clldutils.misc import lazyproperty
from clldutils import jsonlib
from pyglottolog.languoids import Glottocode

from segments import Tokenizer, Profile
from segments.tree import Tree

from cldfbench.dataset import Dataset as BaseDataset
from cldfbench.cldf import CLDFSpec

from pylexibank.util import jsondump, log_dump
from pylexibank import cldf
from pylexibank import models
from pylexibank import transcription
from pylexibank import metadata
from pylexibank import forms
from pylexibank import report

__all__ = ['Dataset', 'ENTRY_POINT']
ENTRY_POINT = 'lexibank.dataset'


class Dataset(BaseDataset):
    """
    A lexibank dataset.

    This object provides access to a dataset's
    - language list as attribute `languages`
    - concept list as attribute `concepts`
    - concepticon concept-list ID as attribute `conceptlist`
    """
    metadata_cls = metadata.LexibankMetadata

    lexeme_class = models.Lexeme
    cognate_class = models.Cognate
    language_class = models.Language
    concept_class = models.Concept

    form_spec = forms.FormSpec()

    @property
    def stats(self):
        if self.dir.joinpath('README.json').exists():
            return jsonlib.load(self.dir / 'README.json')
        return {}

    def __init__(self, concepticon=None, glottolog=None):
        super().__init__()
        if self.__class__ != Dataset:
            if not self.id:
                raise ValueError(
                    "Dataset.id needs to be specified in subclass for %s!" % self.__class__)
        self.unmapped = Unmapped()
        self._json = self.dir.joinpath('lexibank.json')

        self.conceptlists = []
        self.glottolog = glottolog
        self.concepticon = concepticon
        self.tr_analyses = {}
        self.tr_bad_words = []
        self.tr_invalid_words = []

    def cldf_specs(self):
        return CLDFSpec(
            module='Wordlist',
            writer_cls=cldf.LexibankWriter,
            dir=self.cldf_dir,
            metadata_fname=cldf.MD_NAME,
            default_metadata_path=pathlib.Path(__file__).parent / cldf.MD_NAME)

    def _iter_etc(self, what):
        delimiter = '\t'
        path = self.etc_dir / (what + '.tsv')
        if not path.exists():
            delimiter = ','
            path = path.parent / (what + '.csv')
        return reader(path, dicts=True, delimiter=delimiter) if path.exists() else []

    def read_json(self):  # pragma: no cover
        return jsonlib.load(self._json) if self._json.exists() else {}

    def write_json(self, obj):  # pragma: no cover
        jsondump(obj, self._json)

    @lazyproperty
    def sources(self):
        return list(self._iter_etc('sources'))

    @lazyproperty
    def concepts(self):
        return list(self._iter_etc('concepts'))

    @lazyproperty
    def languages(self):
        res = []
        for item in self._iter_etc('languages'):
            if item.get('GLOTTOCODE', None) and not \
                    Glottocode.pattern.match(item['GLOTTOCODE']):  # pragma: no cover
                raise ValueError(
                    "Invalid glottocode {0}".format(item['GLOTTOCODE']))
            res.append(item)
        return res

    @lazyproperty
    def lexemes(self):
        res = {}
        for item in self._iter_etc('lexemes'):
            res[item['LEXEME']] = item['REPLACEMENT']
        return res

    # ---------------------------------------------------------------
    # handling of lexemes/forms/words
    # ---------------------------------------------------------------
    @lazyproperty
    def tokenizer(self):
        """
        Datasets can provide support for segmentation (aka tokenization) in two ways:
        - by providing an orthography profile at etc/orthography.tsv or
        - by overwriting this method to return a custom tokenizer callable.

        :return: A callable to do segmentation.

        The expected signature of the callable is

            def t(item, string, **kw)

        where
        - `item` is a `dict` representing the complete CLDF FormTable row
        - `string` is the string to be segmented
        - `kw` may be used to pass any context info to the tokenizer, when called
          explicitly.
        """
        def get_tokenizer(profile_):
            profile_ = Profile.from_file(str(profile_), form='NFC')
            default_spec = list(next(iter(profile_.graphemes.values())).keys())
            for grapheme in ['^', '$']:
                if grapheme not in profile_.graphemes:
                    profile_.graphemes[grapheme] = {k: None for k in default_spec}
            profile_.tree = Tree(list(profile_.graphemes.keys()))
            return Tokenizer(profile=profile_, errors_replace=lambda c: '<{0}>'.format(c))

        tokenizers = {}
        profile = self.etc_dir / 'orthography.tsv'
        profile_dir = self.etc_dir / 'orthography'
        if profile.exists():
            tokenizers[None] = get_tokenizer(profile)
        if profile_dir.exists() and profile_dir.is_dir():
            for profile in profile_dir.glob('*.tsv'):
                tokenizers[profile.stem] = get_tokenizer(profile)

        if tokenizers:
            def _tokenizer(item, string, **kw):
                kw.setdefault("column", "IPA")
                kw.setdefault("separator", " + ")
                profile = kw.pop('profile', None)
                if profile:
                    tokenizer = tokenizers[profile]
                elif isinstance(item, dict) \
                        and 'Language_ID' in item \
                        and item['Language_ID'] in tokenizers:
                    tokenizer = tokenizers[item['Language_ID']]
                else:
                    tokenizer = tokenizers[None]
                return tokenizer(unicodedata.normalize('NFC', '^' + string + '$'), **kw).split()
            return _tokenizer

    def _cmd_makecldf(self, args):
        # Inject the appropriate CLDFWriter instance:
        self.unmapped.clear()
        self.tr_analyses = {}
        self.tr_bad_words = []
        self.tr_invalid_words = []

        if len(self.metadata.conceptlist):
            self.conceptlists = [
                self.concepticon.conceptlists[key] for key in self.metadata.conceptlist]

        super()._cmd_makecldf(args)

        if args.verbose:
            self.unmapped.pprint()
        assert self.cldf_reader().validate(args.log)

        stats = transcription.Stats(
            bad_words=sorted(self.tr_bad_words[:100], key=lambda x: x['ID']),
            bad_words_count=len(self.tr_bad_words),
            invalid_words=sorted(self.tr_invalid_words[:100], key=lambda x: x['ID']),
            invalid_words_count=len(self.tr_invalid_words))
        for lid, analysis in self.tr_analyses.items():
            for attribute in ['segments', 'bipa_errors', 'sclass_errors', 'replacements']:
                getattr(stats, attribute).update(getattr(analysis, attribute))
            stats.general_errors += analysis.general_errors
            stats.inventory_size += len(analysis.segments) / len(self.tr_analyses)

        error_segments = stats.bipa_errors.union(stats.sclass_errors)
        for i, row in enumerate(stats.bad_words):
            analyzed_segments = []
            for s in row['Segments']:
                analyzed_segments.append('<s> %s </s>' % s if s in error_segments else s)
            stats.bad_words[i] = [
                row['ID'],
                row['Language_ID'],
                row['Parameter_ID'],
                row['Form'],
                ' '.join(analyzed_segments)]

        for i, row in enumerate(stats.invalid_words):
            stats.invalid_words[i] = [
                row['ID'],
                row['Language_ID'],
                row['Parameter_ID'],
                row['Form']]
        # Aggregate transcription analysis results ...
        tr = dict(
            by_language={k: attr.asdict(v) for k, v in self.tr_analyses.items()},
            stats=attr.asdict(stats))

        jsondump(tr, self.cldf_dir / '.transcription-report.json', log=args.log)

        # ... and write a report:
        args.tr_analysis = tr
        self._cmd_readme(args)
        (self.dir / 'TRANSCRIPTION.md').write_text(transcription.report(tr), encoding='utf8')
        log_dump(self.dir / 'TRANSCRIPTION.md', args.log)

    def cmd_readme(self, args):
        res = self.metadata.markdown()
        tr = self.cldf_dir / '.transcription-report.json'
        tr = jsonlib.load(tr) if tr.exists() else None
        res += report.report(self, tr, getattr(args, 'glottolog', None), args.log)
        self.dir.write('FORMS.md', self.form_spec.as_markdown())
        return res


class NonSplittingDataset(Dataset):
    form_spec = forms.FirstFormOnlySpec()


class Unmapped(object):
    def __init__(self):
        self.languages = set()
        self.concepts = set()

    def clear(self):
        self.languages = set()
        self.concepts = set()

    def add_concept(self, **kw):
        self.concepts.add(models.Concept(**kw))

    def add_language(self, **kw):
        self.languages.add(models.Language(**kw))

    @staticmethod
    def quote(v):
        v = '{0}'.format(v or '')
        if ',' in v or len(v.split()) > 1:
            v = '"%s"' % v.replace('"', '""')
        return v

    def pprint(self):
        for objs, cls in [(self.languages, models.Language), (self.concepts, models.Concept)]:
            if objs:
                print('=== Unmapped %ss ===' % cls.__name__)
                print(','.join([a.name.upper() for a in attr.fields(cls)]))
                for row in sorted(map(attr.astuple, objs)):
                    print(','.join(map(self.quote, row)))



"""
List lexibank datasets - installed or loaded into the db.
"""
import collections
import textwrap

import termcolor
from clldutils.markup import Table
from clldutils import licenses
from cldfbench.cli_util import get_datasets

from pylexibank.cli_util import add_db, get_db, add_dataset_spec

COLS = [
    'version',
    'location',
    'license',
    'changes',
    'all_lexemes',
    'lexemes',
    'concepts',
    'languages',
    'families',
    'varieties',
    'macroareas',
]


def register(parser):
    add_db(parser)
    add_dataset_spec(parser, multiple=True)
    parser.add_argument(
        '--all',
        help='Show all columns',
        action='store_true',
        default=False)
    for col in COLS:
        parser.add_argument(
            '--{0}'.format(col),
            help='Show column {0}'.format(col),
            action='store_true',
            default=False)


def run(args):
    db = get_db(args)
    in_db = {r[0]: r[1] for r in db.fetchall('select id, version from dataset')}

    table = Table('ID', 'Title')
    cols = collections.OrderedDict([
        (col, {}) for col in COLS if getattr(args, col, None) or args.all])
    tl = 40
    if cols:
        tl = 25
        table.columns.extend(col.capitalize() for col in cols)

    for col, sql in [
        ('languages', 'glottocodes_by_dataset'),
        ('concepts', 'conceptsets_by_dataset'),
        ('lexemes', 'mapped_lexemes_by_dataset'),
        ('all_lexemes', 'lexemes_by_dataset'),
        ('macroareas', 'macroareas_by_dataset'),
        ('families', 'families_by_dataset'),
    ]:
        if col in cols:
            cols[col] = {r[0]: r[1] for r in db.fetchall(sql)}
    datasets = get_datasets(args)
    for ds in datasets:
        row = [
            termcolor.colored(ds.id, 'green' if ds.id in in_db else 'red'),
            textwrap.shorten(ds.metadata.title or '', width=tl),
        ]
        for col in cols:
            if col == 'version':
                row.append(ds.repo.hash())
            elif col == 'location':
                row.append(termcolor.colored(str(ds.dir), 'green'))
            elif col == 'changes':
                row.append(ds.repo.is_dirty())
            elif col == 'license':
                lic = licenses.find(ds.metadata.license or '')
                row.append(lic.id if lic else ds.metadata.license)
            elif col in ['languages', 'concepts', 'lexemes', 'all_lexemes', 'families']:
                row.append(float(cols[col].get(ds.id, 0)))
            elif col == 'macroareas':
                row.append(', '.join(sorted((cols[col].get(ds.id) or '').split(','))))
            else:
                row.append('')

        table.append(row)
    totals = ['zztotal', len(datasets)]
    for i, col in enumerate(cols):
        if col in ['lexemes', 'all_lexemes']:
            totals.append(sum([r[i + 2] for r in table]))
        elif col == 'languages':
            totals.append(float(db.fetchone(
                "SELECT count(distinct glottocode) FROM languagetable")[0]))
        elif col == 'concepts':
            totals.append(float(db.fetchone(
                "SELECT count(distinct concepticon_id) FROM parametertable")[0]))
        elif col == 'families':
            totals.append(float(db.fetchone(
                "SELECT count(distinct family) FROM languagetable")[0]))
        else:
            totals.append('')
    table.append(totals)
    print(table.render(
        tablefmt='simple', sortkey=lambda r: r[0], condensed=False, floatfmt=',.0f'))
